<?php

namespace Drupal\moon_phases\Controller;

use Drupal\Component\Datetime\DateTimePlus;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Url;
use Drupal\moon_phases\MoonCalc;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Controller for show moon phase.
 */
class MoonPhasesController extends ControllerBase {

  /**
   * The moon phase object.
   *
   * @var \Drupal\moon_phases\MoonCalc
   */
  protected $moon;

  /**
   * An array containing the data pertaining to a phase of the moon.
   *
   * @var array
   */
  protected $phaseData = [];

  /**
   * The current date object.
   *
   * @var \Drupal\Component\Datetime\DateTimePlus
   */
  protected $currentDate;

  /**
   * The description of the moon phase.
   *
   * @var string
   */
  protected $phaseDescription;

  /**
   * The Moon Phases config settings.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $moonConfigSettings;

  /**
   * Returns the renderer service.
   *
   * @var \Drupal\Core\Render\Renderer
   */
  protected $renderer;

  /**
   * Returns the date.formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * Returns the datetime.time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $timeService;

  /**
   * Constructs an MoonPhasesController object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Defines the interface for a configuration object factory.
   * @param \Drupal\Core\Render\Renderer $renderer
   *   Turns a render array into a HTML string.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   Provides an interface defining a date formatter.
   * @param \Drupal\Component\Datetime\TimeInterface $time_service
   *   Defines an interface for obtaining system time.
   */
  public function __construct(ConfigFactoryInterface $config_factory, Renderer $renderer, DateFormatterInterface $date_formatter, TimeInterface $time_service) {
    $this->moonConfigSettings = $config_factory->get('moon_phases.settings');
    $this->renderer = $renderer;
    $this->dateFormatter = $date_formatter;
    $this->timeService = $time_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('renderer'),
      $container->get('date.formatter'),
      $container->get('datetime.time')
    );
  }

  /**
   * Creates the default Moon Phase page.
   *
   * @return array
   *   Returns rendered HTML.
   */
  public function content($date) {
    $this->setCurrentDate($date);
    $this->moon = new MoonCalc($this->currentDate);
    $this->setPhaseData();
    $this->setPhaseDescription();

    return $this->buildPage();
  }

  /**
   * Collates all of the Moon Phase data and returns the rendered page.
   *
   * @return array
   *   Returns a render array.
   */
  public function buildPage() {
    $description = $this->phaseDescription;
    $attribution_url = Url::fromUri('https://svs.gsfc.nasa.gov', [
      'attributes' => [
        'class' => [
          'attribution',
        ],
        'target' => '_blank',
      ],
    ]);

    $attribution = [
      '#title' => $this->t("Images courtesy of NASA's Scientific Visualization Studio"),
      '#type' => 'link',
      '#url' => $attribution_url,
    ];

    $items = [];

    foreach ($this->phaseData as $key => $days) {
      if ($key == 'illumination') {
        $items[] = $this->t('%illum% illumination', ['%illum' => $days]);
      }
      else {
        $items[] = $this->t('%days days until next %phase',
            ['%days' => $days, '%phase' => $key]);
      }
    }

    // Get an unorderd list.
    $item_list = [
      '#theme' => 'item_list',
      '#list_type' => 'ul',
      '#items' => $items,
      '#title' => '',
      '#attributes' => [
        'class' => ['moon-phase-data'],
      ],
    ];

    $build = [
      '#type' => 'moon',
      '#theme' => 'moon',
      '#image' => [
        '#theme' => 'image',
        '#uri' => $this->moon->getImageUri(),
        '#alt' => $this->moon->getMoonPhaseName(),
        '#title' => $this->moon->getMoonPhaseName(),
        '#attributes' => [
          'class' => [
            'moon',
          ],
          'data-attribution' => [
            $this->t("NASA's Scientific Visualization Studio"),
          ],
        ],
      ],
      '#description' => [
        '#markup' => $description,
      ],
      '#phase_data' => [
        '#markup' => $this->renderer->render($item_list),
      ],
      '#previous' => [
        '#title' => $this->t('Previous'),
        '#type' => 'link',
        '#url' => $this->getPreviousLink(),
      ],
      '#next' => [
        '#title' => $this->t('Next'),
        '#type' => 'link',
        '#url' => $this->getNextLink(),
      ],
      '#attribution' => $this->showAttribution() ? $attribution : [],
      '#attributes' => [
        'class' => [
          strtolower(str_replace(' ', '-', $this->moon->getMoonPhaseName())),
        ],
      ],
    ];

    return [
      '#title' => $this->currentDate->format('F jS, Y') . ': ' . $this->moon->getMoonPhaseName(),
      '#markup' => $this->renderer->render($build),
    ];
  }

  /**
   * Sets the Moon Phase data and set the $phaseData array.
   */
  public function setPhaseData() {
    $phases = [
      'New Moon',
      'First Quarter',
      'Full Moon',
      'Last Quarter',
    ];
    foreach ($phases as $phase) {
      $this->phaseData[$phase] = $this->moon->getDaysUntilNextMoonType($phase);
    }
    $this->phaseData['illumination'] = $this->moon->getPercentOfIllumination();
  }

  /**
   * Returns the previous day Url object.
   *
   * @return \Drupal\Core\Url
   *   Returns the previous day Url object.
   *
   * @throws \Exception
   * @throws \InvalidArgumentException
   */
  public function getPreviousLink() {
    $previous_day = new DateTimePlus($this->moon->getMoonPhaseDateTime()->format('Y-m-d'));
    $url = Url::fromRoute('moon_phases.content', ['date' => $previous_day->modify('-1 day')->format('Y-m-d')], [
      'attributes' => [
        'class' => [
          'moon-prev',
        ],
      ],
    ]);
    return $url;
  }

  /**
   * Returns the next day Url object.
   *
   * @return \Drupal\Core\Url
   *   Returns the next day Url object.
   *
   * @throws \Exception
   * @throws \InvalidArgumentException
   */
  public function getNextLink() {
    $next_day = new DateTimePlus($this->moon->getMoonPhaseDateTime()->format('Y-m-d'));
    $url = Url::fromRoute('moon_phases.content', ['date' => $next_day->modify('+1 day')->format('Y-m-d')], [
      'attributes' => [
        'class' => [
          'moon-next',
        ],
      ],
    ]);
    return $url;
  }

  /**
   * Checks whether or not to show the image attribution.
   *
   * @return bool
   *   Returns 1 to show the attribution, else 0.
   */
  public function showAttribution() {
    return $this->moonConfigSettings->get('show_attribution');
  }

  /**
   * Set the description, either from the settings or the default.
   */
  public function setPhaseDescription() {
    $name = strtolower(str_replace(' ', '_', $this->moon->getMoonPhaseName()));
    $this->phaseDescription = $this->moonConfigSettings->get($name)['value'] ?? '';
  }

  /**
   * Set the DateTimePlus object to use for the moon phase.
   *
   * @param string $date
   *   The date to set.
   *
   * @throws \Exception
   */
  private function setCurrentDate($date) {
    $current_time = $this->timeService->getCurrentTime();
    $set_date = $date ? $date . ' ' . $this->dateFormatter->format($current_time, 'custom', ' H:i:s') : $this->dateFormatter->format($current_time, 'custom', 'Y-m-d H:i:s');
    $this->currentDate = new DateTimePlus($set_date);
  }

}
