<?php

namespace Drupal\moon_phases\Controller;

use Drupal\Component\Datetime\DateTimePlus;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Url;
use Drupal\moon_phases\MoonCalc;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Controller for show moon phase.
 */
class MoonPhasesMonthlyController extends ControllerBase {

  /**
   * The moon phase object.
   *
   * @var \Drupal\moon_phases\MoonCalc
   */
  protected $moon;

  /**
   * The starting day of the month date object.
   *
   * @var \Drupal\Component\Datetime\DateTimePlus
   */
  protected $monthStart;

  /**
   * The ending day of the month date object.
   *
   * @var \Drupal\Component\Datetime\DateTimePlus
   */
  protected $monthEnd;

  /**
   * The previous month name.
   *
   * @var string
   */
  protected $previousMonth;

  /**
   * The next month name.
   *
   * @var string
   */
  protected $nextMonth;

  /**
   * The date object to use for iteration.
   *
   * @var \Drupal\Component\Datetime\DateTimePlus
   */
  protected $iterationDate;

  /**
   * An array containing the moons for a given month.
   *
   * @var array
   */
  protected $monthMoons = [];

  /**
   * The Moon Phases config settings.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   *   The Moon Phases config settings.
   */
  protected $moonConfigSettings;

  /**
   * Returns the renderer service.
   *
   * @var \Drupal\Core\Render\Renderer
   */
  protected $renderer;

  /**
   * Returns the date.formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * Returns the datetime.time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $timeService;

  /**
   * Constructs an MoonPhasesMonthlyController object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Defines the interface for a configuration object factory.
   * @param \Drupal\Core\Render\Renderer $renderer
   *   Turns a render array into a HTML string.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   *   Provides an interface defining a date formatter.
   * @param \Drupal\Component\Datetime\TimeInterface $time_service
   *   Defines an interface for obtaining system time.
   */
  public function __construct(ConfigFactoryInterface $config_factory, Renderer $renderer, DateFormatterInterface $date_formatter, TimeInterface $time_service) {
    $this->moonConfigSettings = $config_factory->get('moon_phases.settings');
    $this->renderer = $renderer;
    $this->dateFormatter = $date_formatter;
    $this->timeService = $time_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('renderer'),
      $container->get('date.formatter'),
      $container->get('datetime.time')
    );
  }

  /**
   * Creates the Moon Phase calendar page.
   *
   * @return array
   *   Returns calendar render array.
   */
  public function content($month) {
    $this->setMonthStart($month);
    $this->setMonthEnd();
    $this->setIterationDate();

    $this->moon = new MoonCalc($this->monthStart);

    return [
      '#title' => $this->monthStart->format('F Y'),
      '#markup' => $this->buildContent(),
    ];
  }

  /**
   * Builds the render array and returns rendered HTML.
   *
   * @return string
   *   Returns the rendered HTML.
   */
  public function buildContent() {
    $this->getMonthsMoons();
    $this->padCalendar();
    $attribution_url = Url::fromUri('https://svs.gsfc.nasa.gov', [
      'attributes' => [
        'class' => [
          'attribution',
        ],
        'target' => '_blank',
      ],
    ]);

    $attribution = [
      '#title' => $this->t("Images courtesy of NASA's Scientific Visualization Studio"),
      '#type' => 'link',
      '#url' => $attribution_url,
    ];

    $build = [
      '#type' => 'calendar',
      '#theme' => 'moon_calendar',
      '#previous' => [
        '#type' => 'link',
        '#url' => $this->getPreviousLink(),
        '#title' => $this->t('%month', ['%month' => $this->previousMonth]),
        '#attributes' => [
          'class' => 'moon-prev',
        ],
      ],
      '#next' => [
        '#type' => 'link',
        '#url' => $this->getNextLink(),
        '#title' => $this->t('%month', ['%month' => $this->nextMonth]),
        '#attributes' => [
          'class' => 'moon-next',
        ],
      ],
      '#header' => [
        '#markup' => $this->getHeader(),
      ],
      '#calendar' => [
        '#markup' => $this->buildArray(),
      ],
      '#attribution' => $this->showAttribution() ? $attribution : [],
    ];

    return $this->renderer->render($build);
  }

  /**
   * Creates the calendar header.
   *
   * @return string
   *   Returns the rendered HTML.
   */
  public function getHeader() {
    $days = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
    $header = [
      '#type' => 'header',
    ];
    foreach ($days as $day) {
      $header[] = [
        '#markup' => '<div class="' . strtolower($day) . '">' . $day . '</div>',
      ];
    }

    return $this->renderer->render($header);
  }

  /**
   * Gets all of the moons for a given month.
   */
  public function getMonthsMoons() {
    $iterate_date = $this->iterationDate;
    $end_date = $this->getDateAsTimestamp($this->monthEnd);

    while ($this->getDateAsTimestamp($iterate_date) <= $end_date) {
      $moon = new MoonCalc($iterate_date);
      $this->monthMoons[$iterate_date->format('j')] = [
        'phase_name' => $moon->getMoonPhaseName(),
        'image' => $moon->getImageUri(),
        'date' => $iterate_date->format('j'),
        'full_date' => $iterate_date->format('l, F jS, Y'),
        'day_of_week' => $iterate_date->format('w'),
        'weekday' => strtolower($iterate_date->format('l')),
      ];
      $iterate_date->modify('+1 day');
    }
  }

  /**
   * Builds the render array for all of the days of the month.
   *
   * @return string
   *   Returns the rendered HTML.
   */
  private function buildArray() {
    $calendar = [];
    foreach ($this->monthMoons as $day) {
      // If this item is a placeholder, add an empty item to the render array.
      if (array_key_exists('placeholder', $day)) {
        $calendar[] = [
          '#type' => 'moon',
          '#theme' => 'moon_day',
          '#phase_name' => [
            '#markup' => '&nbsp;',
          ],
          '#attributes' => [
            'class' => [
              'moon-calendar-day',
              'placeholder',
            ],
          ],
        ];
      }
      // Add the moon phase data to the render array.
      else {
        $calendar[] = [
          '#type' => 'moon',
          '#theme' => 'moon_day',
          '#image' => [
            '#theme' => 'image',
            '#uri' => $day['image'],
            '#alt' => $day['phase_name'],
            '#title' => $day['phase_name'],
            '#attributes' => [
              'class' => [
                'moon',
              ],
              'data-attribution' => [
                "NASA's Scientific Visualization Studio",
              ],
            ],
          ],
          '#full_date' => [
            '#markup' => $day['full_date'],
          ],
          '#day' => [
            '#markup' => $day['date'],
          ],
          '#phase_name' => [
            '#markup' => $day['phase_name'],
          ],
          '#attributes' => [
            'class' => [
              $day['weekday'],
              'moon-calendar-day',
              'moon-day',
            ],
          ],
        ];
      }
    }

    return $this->renderer->render($calendar);
  }

  /**
   * Pad the beginning of the month for months that start after Sunday.
   */
  private function padCalendar() {
    $first = reset($this->monthMoons);
    if ($first['day_of_week'] > 0) {
      $count = $first['day_of_week'];
      $placeholder = [
        'placeholder' => TRUE,
      ];
      for ($i = $count; $i > 0; $i--) {
        array_unshift($this->monthMoons, $placeholder);
      }
    }
  }

  /**
   * Create the previous day link.
   *
   * @return \Drupal\Core\Url
   *   Returns the URL object for the previous month link.
   *
   * @throws \Exception
   * @throws \InvalidArgumentException
   */
  private function getPreviousLink() {
    $previous_month = new DateTimePlus($this->monthStart->format('Y-m-d'));
    $url = Url::fromRoute('moon_phases.current', ['month' => $previous_month->modify('-1 month')->format('Y-m')], [
      'attributes' => [
        'class' => [
          'moon-prev',
        ],
      ],
    ]);
    $this->previousMonth = $previous_month->format('F');
    return $url;
  }

  /**
   * Create the previous day link.
   *
   * @return \Drupal\Core\Url
   *   Returns the URL object for the next month link.
   *
   * @throws \Exception
   * @throws \InvalidArgumentException
   */
  private function getNextLink() {
    $next_month = new DateTimePlus($this->monthStart->format('Y-m-d'));
    $url = Url::fromRoute('moon_phases.current', ['month' => $next_month->modify('+1 month')->format('Y-m')], [
      'attributes' => [
        'class' => [
          'moon-prev',
        ],
      ],
    ]);
    $this->nextMonth = $next_month->format('F');
    return $url;
  }

  /**
   * Checks whether or not to show the image attribution.
   *
   * @return bool
   *   Returns 1 to show the attribution, else 0.
   */
  private function showAttribution() {
    return $this->moonConfigSettings->get('show_attribution');
  }

  /**
   * Set the start of the month.
   *
   * @param mixed $date
   *   The month to display or null for current month.
   *
   * @throws \Exception
   */
  private function setMonthStart($date) {
    $set_date = $date ?? $this->dateFormatter->format($this->timeService->getCurrentTime(), 'custom', 'Y-m');
    $this->monthStart = new DateTimePlus($set_date);
  }

  /**
   * Set the last day of the month.
   */
  private function setMonthEnd() {
    $month_end = new DateTimePlus($this->monthStart->format('Y-m-d'));
    $this->monthEnd = new DateTimePlus($month_end->modify('last day of this month')->format('Y-m-d'));
  }

  /**
   * Create the \DateTimePlus object to create the days of the month.
   */
  private function setIterationDate() {
    $this->iterationDate = new DateTimePlus($this->monthStart->format('Y-m-d'));
  }

  /**
   * Creates a Unix timestamp for a given DateTimePlus object.
   *
   * @param \Drupal\Component\Datetime\DateTimePlus $date
   *   The date to transform.
   *
   * @return int
   *   Returns the date as a Unix timestamp.
   */
  private function getDateAsTimestamp(DateTimePlus $date) {
    return strtotime($date->format('Y-m-d'));
  }

}
