<?php

namespace Drupal\moon_phases\Plugin\Block;

use Drupal\Component\Datetime\DateTimePlus;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Image\ImageFactory;
use Drupal\Core\Link;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\moon_phases\MoonCalc;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides the Moon Phase block.
 *
 * @Block(
 *  id = "moon_phase_block",
 *  admin_label = @Translation("Moon Phase"),
 *  category = @Translation("Moon Phase"),
 * )
 */
class MoonPhaseBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * Returns the renderer service.
   *
   * @var \Drupal\Core\Render\Renderer
   */
  protected $renderer;

  /**
   * Returns the image.factory service.
   *
   * @var \Drupal\Core\Image\ImageFactory
   */
  protected $imageFactory;

  /**
   * Returns the entity_type.manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs an MoonPhaseBlock object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, Renderer $renderer, ImageFactory $image_factory, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->renderer = $renderer;
    $this->imageFactory = $image_factory;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('renderer'),
      $container->get('image.factory'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'moon_block_items' => [],
      'image_style' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $form = parent::blockForm($form, $form_state);

    $image_styles = $this->entityTypeManager->getStorage('image_style')->loadMultiple();
    $image_styles_list = [];
    foreach ($image_styles as $image_style) {
      $image_styles_list[$image_style->id()] = $image_style->label();
    }

    $form['image_style'] = [
      '#type' => 'select',
      '#title' => $this->t('Image Style'),
      '#description' => $this->t('Select image style that will be used in the block.'),
      '#options' => $image_styles_list,
      '#default_value' => $this->configuration['image_style'] ?? NULL,
      '#empty_option' => $this->t('- Select image style -'),
    ];

    $form['moon_block_items'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Block elements'),
      '#default_value' => $this->configuration['moon_block_items'],
      '#options' => [
        'getDaysUntilNextFullMoon' => $this->t('Days until next Full Moon'),
        'getDaysUntilNextNewMoon' => $this->t('Days until next New Moon'),
        'getPercentOfIllumination' => $this->t('Percentage of illumination'),
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $this->configuration['image_style'] = $form_state->getValue('image_style');
    $this->configuration['moon_block_items'] = $form_state->getValue('moon_block_items');

    parent::blockSubmit($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    $date = new DateTimePlus();
    $moon = new MoonCalc($date);

    $items = [];
    foreach ($this->configuration['moon_block_items'] as $key => $show) {
      if ($show) {
        switch ($key) {
          case 'getDaysUntilNextFullMoon':
            $items[] = $this->t('%days days until the next full moon',
                ['%days' => floor($moon->getDaysUntilNextMoonType('Full Moon'))]);
            break;

          case 'getDaysUntilNextNewMoon':
            $items[] = $this->t('%days days until the next new moon',
                ['%days' => floor($moon->getDaysUntilNextMoonType('New Moon'))]);
            break;

          case 'getPercentOfIllumination':
            $items[] = $this->t('%illum% illuminated',
                ['%illum' => round($moon->getPercentOfIllumination())]);
            break;
        }
      }
    }

    $more_info_url = Url::fromRoute('moon_phases.content', [], [
      'attributes' => [
        'class' => [
          'moon-more-link',
        ],
      ],
    ]);
    $items[] = Link::fromTextAndUrl($this->t('See more'), $more_info_url);

    // Get an unorderd list.
    $item_list = [
      '#theme' => 'item_list',
      '#list_type' => 'ul',
      '#items' => $items,
      '#title' => '',
      '#attributes' => [
        'class' => ['moon-phase-summary'],
      ],
    ];
    $summary = $this->renderer->render($item_list);
    $phase = $moon->getMoonPhaseName();
    $image_uri = $moon->getImageUri();

    $image_build = [
      '#theme' => 'image',
      '#uri' => $image_uri,
      '#alt' => $phase,
      '#title' => $phase,
    ];

    $cache_tag = '';

    // Build an image style if one is selected on the block settings form.
    if (!empty($this->configuration['image_style'])
      && $image_style = $this->entityTypeManager->getStorage('image_style')->load($this->configuration['image_style'])
    ) {
      $image = $this->imageFactory->get($image_uri);
      $destination_uri = $image_style->buildUri($image_uri);
      $image_style->createDerivative($image_uri, $destination_uri);

      $image_build = [
        '#theme' => 'image_style',
        '#style_name' => $this->configuration['image_style'],
        '#alt' => $phase,
        '#title' => $phase,
        '#width' => $image->getWidth(),
        '#height' => $image->getHeight(),
        '#uri' => $destination_uri,
      ];

      $cache_tag = 'config:image.style.' . $image_style->getName();
    }

    $build = [
      '#type' => 'moon',
      '#theme' => 'moon_block',
      '#phase_name' => $phase,
      '#image' => $image_build,
      '#summary' => $summary,
      '#cache' => [
        'contexts' => [
          'timezone',
          'moon_phases_date',
        ],
      ],
    ];

    if ($cache_tag) {
      $build['#cache']['tags'][] = $cache_tag;
    }

    $output = $this->renderer->render($build);

    return [
      '#markup' => $output,
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function blockAccess(AccountInterface $account) {
    return AccessResult::allowedIfHasPermission($account, 'access moon phases');
  }

}
